import {
  Injectable,
  CanActivate,
  ExecutionContext,
  HttpException,
  HttpStatus,
} from '@nestjs/common';
import { Reflector } from '@nestjs/core';
import * as jwt from 'jsonwebtoken';
import { InjectRepository } from '@nestjs/typeorm';
import { Repository } from 'typeorm';

import { Game_PersonEntity } from '../game/game.entity';

@Injectable()
export class RolesGuard implements CanActivate {
  constructor(
    private readonly reflector: Reflector,
    @InjectRepository(Game_PersonEntity)
    private game_PersonRepository: Repository<Game_PersonEntity>,
  ) {}

  async canActivate(context: ExecutionContext): Promise<boolean> {
    // get roles that are allowed access, identified by @Roles('role') decorators in controllers
    const roles = this.reflector.get<string[]>('roles', context.getHandler());
    if (!roles) {
      return true;
    }
    const request = context.switchToHttp().getRequest();
    // check for authorization header
    if (!request.headers.authorization) {
      return false;
    }
    const gameId = request.params.id;
    request.user = await this.getUserObject(request.headers.authorization);
    const role = await this.game_PersonRepository.findOne({
      person: request.user['id'],
      game: gameId,
    });
    // check that the role matches the criteria and that token is valid for this game
    return role && roles.includes(role['role']);
  }

  async getUserObject(auth: string) {
    // check if header contains Bearer
    if (auth.split(' ')[0] !== 'Bearer') {
      throw new HttpException('Invalid token', HttpStatus.FORBIDDEN);
    }
    // get the token
    const token = auth.split(' ')[1];
    try {
      return await jwt.verify(token, process.env.SECRET);
    } catch (err) {
      const message = `Token error: ${err.message || err.name}`;
      throw new HttpException(message, HttpStatus.FORBIDDEN);
    }
  }
}
